local AddonName, GGL = ...

local L = GGL.Locale

local LibDeflate = LibStub:GetLibrary("LibDeflate")

-- lua
local type = type
local tonumber = tonumber
local setmetatable, select = setmetatable, select
local strsub, strtrim, str_match, format, strfind = string.sub, strtrim, string.match, string.format, string.find
local tblremove = table.remove
local bit_band = bit.band

-- wow
local C_TimerAfter = C_Timer.After
local CombatLogGetCurrentEventInfo = CombatLogGetCurrentEventInfo
local UnitIsTapDenied = UnitIsTapDenied

-- other
local PlayerGUID
local ToolTipHookIsInitialized

-- ########################################
-- Init
-- ########################################
local function Init()
	PlayerGUID = UnitGUID("player")

    -- activate tooltip hooks
    ToolTipHookIsInitialized = true
    GGL:UpdateTooltipHooks()
end
GGL:AddInitFuncPlayerLogin(Init)

-- ########################################
-- Base
-- ########################################
local NPC_GUID_DELIMITER = "-"

--- Formats a players display name
-- @param		name		<string>
-- @return	formatedName
function GGL.FormatDisplayName(name)
    name = strtrim(name)
	return strsub(name or "", 1, GGL.MAX_NAME_LENGTH)
end

--- Get NPC'd from a given GUID
-- @param		guid		<string> unit GUID
-- @return	npcID
function GGL.GetNpcIDFromGuid(guid)
    local npcID = select(6, strsplit(NPC_GUID_DELIMITER, guid))
	if npcID then
		return tonumber(npcID)
	end
end

-- ########################################
-- Error log
-- ########################################
function GGL:AddIntoErrorLog(name, data)
    if self.db.ErrorLogEnabled then
        GoldGoblinLootDB.ErrorLog[ #GoldGoblinLootDB.ErrorLog + 1 ] = { date(GGL.ERROR_LOG_DATE_FORMAT), name, data, GGL.GetTime() }
    end
end

--- Get the full error log as string
-- @return logString, numEntrys
function GGL:GetErrorLogAsString()
    local s = ""
    for logNum = #GoldGoblinLootDB.ErrorLog, 1, -1 do
        local log = GoldGoblinLootDB.ErrorLog[logNum]
        s = s..format("%s - %s - ", log[1], log[2])
        if type(log[3]) == "table" then
            s = s.."{ "
            if #log[3] > 0 then
                for k,v in ipairs(log[3]) do
                    s = s..(k>1 and ", " or "")..v
                end
            else
                local c = 0
                for k,v in pairs(log[3]) do
                    s = s..format("%s[%s] = %s", (c>0 and ", " or ""), tostring(k), tostring(v))
                    c = c + 1
                end
            end
            s = s.." }\n"
        else
            s = s..(log[3] or "nil").."\n"
        end
    end
    return s, #GoldGoblinLootDB.ErrorLog
end

function GGL:PurgeErrorLog()
    if GoldGoblinLootDB.ErrorLog then
        wipe(GoldGoblinLootDB.ErrorLog)
    end
end

-- ########################################
-- Fonts
-- ########################################
local FontRegister = {}

function GGL.GetFont(size, flags)
    if FontRegister[size..(flags or "")] then
        return FontRegister[size..(flags or "")]
    else
        local font = CreateFont("GGL-FONT-"..size..(flags or ""))
        font:SetFont(GGL.FONT_PATH, size, flags)
        FontRegister[size..(flags or "")] = font
        return font
    end
end

-- ########################################
-- Archivist
-- ########################################
EnableAddOn("GoldGoblinLootArchive")
local Archivist

function GGL:AddSessionIntoArchive(sessionDB)
    if not IsAddOnLoaded("GoldGoblinLootArchive") then
        local ok, reason = LoadAddOn("GoldGoblinLootArchive")
        if not ok then
            GGL:AddIntoErrorLog("AddSessionIntoArchive", reason)
        end
    end
    if not Archivist and GGL.Archivist and GGL.Archivist.Initialize then
        Archivist = GGL.Archivist:Initialize(GoldGoblinLootArchiveDB)
    elseif not Archivist then
        GGL:AddIntoErrorLog("AddSessionIntoArchive", "Archivist full missing")
    end
    if Archivist and Archivist.Create then
        if Archivist:Check("ReadOnly", sessionDB.sessionID) then
            -- log exists?? we delete it first so we can replace it
            Archivist:Delete("ReadOnly", sessionDB.sessionID, true)
            GGL:AddIntoErrorLog("AddSessionIntoArchive", "Remove double entry from Archivist "..(sessionDB.sessionID or ""))
        end
        GoldGoblinLootArchiveDataDB[sessionDB.sessionID] = {
            startTime = sessionDB.startTime,
            endTime = sessionDB.endTime,
        }
        local store = Archivist:Create("ReadOnly", sessionDB.sessionID, sessionDB)
        Archivist:CloseStore(store)
    else
        GGL:AddIntoErrorLog("AddSessionIntoArchive", "Archivist init missing")
    end
end

-- ########################################
-- GameTooltip hook
-- ########################################
local DEFAULT_HOOK_TOOLTIP = "GameTooltip"
local TooltipHookRegister = {}
local ActiveTooltipHookRegister = {}

--- Adds a hook into the DEFAULT_HOOK_TOOLTIP tooltip
-- Hooks cant be Unhooked so a active check is also needed in the "func" function
-- activeCheck gets only checked when hooking
-- @param       script          <string> ToolTip script
-- @param       func            <function> Function to hook into the tt
-- @param       activeCheck     <function/bool> Check befor hook into tt, nil/true means always hook.
function GGL:TooltipHook(script, func, activeCheck)
    if not func or type(func) ~= "function" or not script then return end
    if not TooltipHookRegister[script] then
        TooltipHookRegister[script] = {}
    end
    if not TooltipHookRegister[script][func] then
        local hookTT = _G[DEFAULT_HOOK_TOOLTIP]
        if hookTT.HookScript then
            TooltipHookRegister[script][func] = activeCheck or (activeCheck == nil and true or false)
            if ToolTipHookIsInitialized then
                GGL:UpdateTooltipHooks()
            end
        end
    end
end

function GGL:TooltipHookSetActive(script, func)
    if TooltipHookRegister[script] and TooltipHookRegister[script][func] == false then
        TooltipHookRegister[script][func] = true
        GGL:UpdateTooltipHooks()
    elseif not TooltipHookRegister[script] then
        GGL:AddIntoErrorLog("GGL:TooltipHookSetActive / script missing", script)
    end
end

--- Update all hooks and add missing hooks
function GGL:UpdateTooltipHooks()
    if not ToolTipHookIsInitialized then return end
    local hookTT = _G[DEFAULT_HOOK_TOOLTIP]
    if not hookTT.HookScript then return end
    for script, scriptData in pairs(TooltipHookRegister) do
        for func, activeCheck in pairs(scriptData) do
            if not ActiveTooltipHookRegister[func] and activeCheck ~= false and ( activeCheck == true or activeCheck() ) then
                hookTT:HookScript(script, func)
                ActiveTooltipHookRegister[func] = true
            end
        end
    end
end


-- ########################################
-- UnitTooltip display
-- ########################################
local TT_HOOK_GUID_TEXT_FORMAT = "|cff33ff99GoldGoblinLoot|r: %s"

local OnTooltipSetUnit_ShowAttendeesInTooltip_IsActive
local ActiveGuidRegister

local function OnTooltipSetUnit_ShowAttendeesInTooltip(self)
    if not ActiveGuidRegister then return end
    if not GGL.db.Settings.ShowAttendeesInTooltip then return end
    if self:IsForbidden() then return end
    local name, target = self:GetUnit()
    if not target then return end
    local guid = UnitGUID(target)
    if guid and ActiveGuidRegister[guid] then
        self:AddLine(format(TT_HOOK_GUID_TEXT_FORMAT, ActiveGuidRegister[guid].displayName))
    end
end

GGL:TooltipHook("OnTooltipSetUnit", OnTooltipSetUnit_ShowAttendeesInTooltip, false)

--- Sets the active table for unitGUIDS to show a text
-- @param       register        <table> { [unitGUID] = playerTemplate }
function GGL:SetActiveGuidRegisterForTooltip(register)
    ActiveGuidRegister = register
    if not OnTooltipSetUnit_ShowAttendeesInTooltip_IsActive then
        GGL:TooltipHookSetActive("OnTooltipSetUnit", OnTooltipSetUnit_ShowAttendeesInTooltip)
        OnTooltipSetUnit_ShowAttendeesInTooltip_IsActive = true
    end
end

--- Gets the active table for unitGUIDS to show a text
function GGL:GetActiveGuidRegisterForTooltip()
    return ActiveGuidRegister
end


-- ########################################
-- Settings Import/Export
-- ########################################

-- local compress_deflate = LibDeflate:CompressDeflate(example_input)
-- local decompress_deflate = LibDeflate:DecompressDeflate(compress_deflate)

function GGL:OpenSetSettingsExport(setID, exportName)
    local set = GGL.db.Config.Sets[GGL.SET_ID_LIST.Set][setID]
    if not set then return end

    local passSetD, setData = GGL:Deserialize(set.setting)

    if not passSetD then return end

    local serTab = {
        exportName = exportName or "i: "..(set.name or "???"),
        id = "i"..GGL.GetTime()
    }
    local setsDbTable = GGL.db.Config.Sets

    for setDesc, setKey in pairs(GGL.SET_ID_LIST) do
        if setDesc ~= "Set" then
            if setData[setKey] == true then
                serTab[setKey] = true
            elseif setData[setKey] and setsDbTable[setKey][ setData[setKey] ] then
                serTab[setKey] = setsDbTable[setKey][ setData[setKey] ].setting
            else
                -- fallback to base
                serTab[setKey] = setsDbTable[setKey].Base.setting
            end
        end
    end

    local ecodedString = GGL:Serialize(serTab)
    local compressed = LibDeflate:CompressDeflate(ecodedString)
	local encoded = LibDeflate:EncodeForPrint(compressed)

    GGL.Export:FillExportWindow(encoded, "text")
    return encoded
end

function GGL:ImportSetSettings(importString)
    if not importString then return end
    local compressed = LibDeflate:DecodeForPrint(importString)
    local serialized = LibDeflate:DecompressDeflate(compressed)
    local pass, data = GGL:Deserialize(serialized)

    if not pass then return end

    local setsDbTable = GGL.db.Config.Sets

    local exportName, exportID = data.exportName, data.id
    data.exportName, data.id = nil, nil
    -- add set
    setsDbTable[GGL.SET_ID_LIST.Set][exportID] = {
        name = exportName
    }
    -- table for set settings
    local setSettingsTab = {}
    -- fill config
    for setKey, settingString in pairs(data) do
        if settingString ~= true then
            setSettingsTab[setKey] = exportID
            setsDbTable[setKey][exportID] = {
                name = exportName,
                setting = settingString
            }
        else
            setSettingsTab[setKey] = true
        end
    end
    -- set set setting
    setsDbTable[GGL.SET_ID_LIST.Set][exportID].setting = GGL:Serialize(setSettingsTab)
    return exportID
end

-- ########################################
-- LootMob (FastLoot)
-- ########################################
local LootMob_LoopRunning = false
local function LootMob_LootLoop()
    local numLoot = GetNumLootItems()
	if numLoot > 0 then
        for slot = 1, numLoot do
            LootSlot(slot)
            numLoot = numLoot - 1
        end
    end

    if numLoot > 1 then
        if not LootMob_LoopRunning then
            LootMob_LoopRunning = true
            C_TimerAfter(0.01, LootMob_LootLoop)
        end
    else
        LootMob_LoopRunning = false
        CloseLoot()
    end
end

--- Starts looting the open loot window
function GGL:LootMob()
    if GGL.db.Settings.FastAutoLoot and not LootMob_LoopRunning then
        LootMob_LootLoop()
    end
end

-- ########################################
-- DestroyGrayBagItems
-- ########################################
local MerchantIsOpen = false

--- Destroys the first grey item in the bag
-- @param       silent      <bool> print out chat msg
function GGL:DestroyGrayBagItems(silent)
    for bag = 0, NUM_BAG_SLOTS do
        for slot = 1, GetContainerNumSlots(bag) do
            local icon, itemCount, locked, quality, readable, lootable, itemLink, isFiltered, noValue, itemID = GetContainerItemInfo(bag, slot)
			if itemID and not locked and quality == 0 then
                if MerchantIsOpen then
                    UseContainerItem(bag, slot)
                else
                    PickupContainerItem(bag, slot)
                    DeleteCursorItem()
                    if GGL.db.Settings.ShowDestroyMessage then
                        GGL:FormatPrint(L["%d x %s destroyed."], itemCount, itemLink)
                    end
                    return -- with 9.1.5 only one item can be destroyed after a HW event
                end
			end
        end
    end
--[[
    if not silent and counter > 0 then
        GGL:FormatPrint(L["%d gray items destroyed."], counter)
    end

    return counter
]]--
end

function GGL:MERCHANT_SHOW()
    MerchantIsOpen = true
end

function GGL:MERCHANT_CLOSED()
    MerchantIsOpen = false
end

GGL:RegisterEvent("MERCHANT_SHOW")
GGL:RegisterEvent("MERCHANT_CLOSED")

-- ########################################
-- Timer
-- ########################################
local TIMER_PROTO = {
    --- Starts a timer if its not registered
    -- @param       interval        <number> in seconds
    -- @param       callback        <function>
    ["After"] = function(self, interval, callback)
        if interval and callback and not self.register[callback] then
            self.register[callback] = true
            C_TimerAfter(interval, callback)
        end
    end,

    --- This starts a new timer that runs as long the callback returns true
    -- @param       interval        <number> in seconds
    -- @param       callback        <function> resumes timer as long it returns true
    -- @callback    elapsed         <number> calls * interval
    ["ManagedAfter"] = function(self, interval, callback)
        if interval and callback and not self.register[callback] then
            local calls = 0
            local function managedAfter()
                calls = calls + 1
                if callback(calls*interval) then
                    C_TimerAfter(interval, managedAfter)
                else
                    self:Unregister(managedAfter)
                end
            end
            C_TimerAfter(interval, managedAfter)
            self.register[managedAfter] = callback
        end
    end,

    --- Unregister a callback from the list so it can set again
    -- @param       callback        <function>
    ["Unregister"] = function(self, callback)
        if callback and self.register[callback] then
            self.register[callback] = nil
            return true
        end
        return false
    end,
}
local TIMER_PROTO_MT = {__index = TIMER_PROTO}

--- Returns a table with an enhanced C_Timer
--- See TIMER_PROTO for more info
function GGL:GetTimer()
    local timer = {
        register = {}
    }

    setmetatable(timer, TIMER_PROTO_MT)

    return timer
end

-- ########################################
-- CachTable
-- ########################################
local EMPTY_TABLE = {}
local WEAK_TABLE_MT = {__mode = "kv"}
local WEAK_TABLE_INDEX_MT = {__mode = "kv", __index = function(k,v) return EMPTY_TABLE end }

--- Gets a cache table
function GGL:GetCacheTable()
    return setmetatable({}, WEAK_TABLE_MT)
end

--- Gets a cache table
function GGL:GetCacheTableWithEmptyTableIndex()
    return setmetatable({}, WEAK_TABLE_INDEX_MT)
end

-- ########################################
-- ItemMouseFunctions
-- ########################################

--- Shows a item in DressUp
-- @param       item        <item> itemID or itemLink
function GGL:ItemMouseShowInDressUp(item)
    if not item then return end
    if type(item) == "number" then
        local _
        _, item = GetItemInfo(item)
    end
    if not item then return end

    return DressUpItemLink(item)
end

--- Add a item into chat editbox
-- @param       item        <item> itemID or itemLink
function GGL:ItemMouseShowInChatEdit(item)
    if not item then return end
    local _
    _, item = GetItemInfo(item)
    if not item then return end
    _G.ChatEdit_InsertLink(item)
end

--- Selects the right function on click
-- @param       item        <item> itemID or itemLink
-- @param       button      <string> mouse button
function GGL:ItemMouseOnClick(item, buttonFrame, button)
    if not item then return end
    if IsControlKeyDown() then
        GGL:ItemMouseShowInDressUp(item)
    elseif IsShiftKeyDown() then
        GGL:ItemMouseShowInChatEdit(item)
    elseif IsAltKeyDown() then
        if type(item) == "string" then
            item = str_match(item, "item:(%d+)")
        end
        GGL:TUJCopyBox_Open(buttonFrame, item)
        --GGL:WoWHeadCopyBox_Open(buttonFrame, nil, item)
    end
end

-- ########################################
-- CopyBox
-- ########################################
local WOW_HEAD_LINK, WOW_HEAD_LINK_LOC = "https://www.wowhead.com/%s=%d", "https://%s.wowhead.com/%s=%d"
local WOW_HEAD_LOCALE
local TUJ_LINK = "https://theunderminejournal.com/#%s/%s/item/%d"
local TUJ_REGION, TUJ_REALM

local function InitCopyBox()
    -- locale for WoWHead
    local locale = GetLocale()
    if locale == "deDE" then WOW_HEAD_LOCALE = "de"
    elseif locale == "esMX" then WOW_HEAD_LOCALE = "es"
    elseif locale == "esES" then WOW_HEAD_LOCALE = "es"
    elseif locale == "frFR" then WOW_HEAD_LOCALE = "fr"
    elseif locale == "itIT" then WOW_HEAD_LOCALE = "it"
    elseif locale == "ptBR" then WOW_HEAD_LOCALE = "pt"
    elseif locale == "ruRU" then WOW_HEAD_LOCALE = "ru"
    elseif locale == "koKR" then WOW_HEAD_LOCALE = "ko"
    elseif locale == "zhCN" then WOW_HEAD_LOCALE = "cn"
    elseif locale == "zhTW" then WOW_HEAD_LOCALE = "cn" end

    -- Region for TUJ
    local regionID = GetCurrentRegion()
    if regionID == 1 then TUJ_REGION = "us"
    elseif regionID == 2 then TUJ_REGION = "kr"
    elseif regionID == 3 then TUJ_REGION = "eu"
    elseif regionID == 4 then TUJ_REGION = "tw"
    elseif regionID == 5 then TUJ_REGION = "us" end -- China :s
    -- Relamname for TUJ
    TUJ_REALM = GetNormalizedRealmName()

    InitCopyBox = nil
end

local CopyBox_Frame
local COPY_BOX_HIDE_AFTER, COPY_BOX_HIDE_AFTER_ENTER = 3, 0.5

local function CopyBox_SetCopyText(self, text)
	if not text or text == self.text then
		self.text = nil
		self:Hide()
		return false
	end

	if self:IsShown() then self:Hide() end
	self:SetText(text)
	self.tLenght:SetText(text)
	self:SetWidth(self.tLenght:GetStringWidth() + 20)
	self.text = text

	self:Show()
	return true
end

local function CopyBox_OnUpdate(self, elapsed)
	if not self.time then return end
	self.time = self.time - elapsed
	if self.time <= 0 then
		self:Hide()
	end
end

local function CopyBox_Create()
	local frame = CreateFrame("EditBox")
	frame:ClearAllPoints()
	frame:SetPoint("TOPLEFT", 70, 4)
	frame:SetHeight(16)
	frame:SetFontObject("GameFontNormal")
	frame:SetBlinkSpeed(0)
	frame:SetAutoFocus(false)
	frame:EnableKeyboard(false)
	frame:SetScript("OnKeyDown", function() end)
	frame:SetScript("OnMouseUp", function()
		if frame:IsMouseOver() then
			frame:HighlightText()
		else
			frame:HighlightText(0, 0)
		end
	end)
	frame:SetScript("OnEnter", function(self)
		local tooltip = _G.GameTooltip
		tooltip:SetOwner(self, "ANCHOR_BOTTOM", 0, -5)
		tooltip:SetText(L["Ctrl + C to copy"], nil, nil, nil, nil, true)
		tooltip:Show()
		self:HighlightText()
		self:SetFocus()
		self.time = nil
	end)
	frame:SetScript("OnLeave", function(self)
		_G.GameTooltip:Hide()
		self:ClearFocus()
		self.time = COPY_BOX_HIDE_AFTER_ENTER -- let it stay some time
	end)
	frame:SetScript("OnShow", function(self)
		self.time = COPY_BOX_HIDE_AFTER
	end)
	frame:SetScript("OnHide", function(self)
		if self:IsShown() then self:Hide() end -- fix for when the frame itself is not hidden
		_G.GameTooltip:Hide()
		self:ClearFocus()
		self.text = nil
		self.time = nil
	end)
	frame:SetScript("OnUpdate", CopyBox_OnUpdate)

	frame.bg = frame:CreateTexture(nil, "BACKGROUND")
	frame.bg:SetAllPoints(frame)
	frame.bg:SetColorTexture(0.1, 0.1, 0.1, 1.0)

	frame.tLenght = frame:CreateFontString(nil, 'ARTWORK', 'GameFontNormal')
	frame.tLenght:Hide()

	frame.SetCopyText = CopyBox_SetCopyText

	return frame
end

function GGL:CopyBox_Show(button, text)
	if not CopyBox_Frame then
		CopyBox_Frame = CopyBox_Create()
	end
	if CopyBox_Frame:SetCopyText(text) then
		CopyBox_Frame:ClearAllPoints()
		CopyBox_Frame:SetParent(button)
		CopyBox_Frame:SetPoint("TOPLEFT", button, "BOTTOMLEFT", 0, -1)
	end
end

function GGL:CopyBox_Hide()
	if CopyBox_Frame and CopyBox_Frame:IsShown() then
		CopyBox_Frame:Hide()
	end
end

--- Opens a small box with an wow head link
-- @param       button          <buttonFrame>
-- @param       type            <string> WoWHead link type (item, npc, ...)
-- @param       id              <number> id
function GGL:WoWHeadCopyBox_Open(button, type, id)
    if InitCopyBox then
        InitCopyBox()
    end
    if id then
        type = type or "item"
		if not WOW_HEAD_LOCALE then
			GGL:CopyBox_Show(button, format(WOW_HEAD_LINK, type, id))
		else
			GGL:CopyBox_Show(button, format(WOW_HEAD_LINK_LOC, WOW_HEAD_LOCALE, type, id))
		end
	end
end

--- Opens a small box with an TUJ link
-- @param       button          <buttonFrame>
-- @param       id              <number> id
function GGL:TUJCopyBox_Open(button, itemID)
    if InitCopyBox then
        InitCopyBox()
    end
    GGL:CopyBox_Show(button, format(TUJ_LINK, TUJ_REGION, TUJ_REALM, itemID))
end

-- ########################################
-- ItemLink
-- #########################################
local ItemLinkCache = GGL:GetCacheTable()
local ITEM_LOG_FORMAT = "(%d+),*(.*)"
local ITEM_LINK_FORMT = "item:%d::::::::::::%s"
--/run print(GetItemInfo("item:3792::::::::::::1:6656:2:28:217:9:30:::"))
--- Formats item data get from Loot module
-- @return itemID, itemString
function GGL:GetItemDataFromItemLog(itemData)
    if not itemData then return end
    if ItemLinkCache[itemData] then
        return ItemLinkCache[itemData][1], ItemLinkCache[itemData][2]
    elseif type(itemData) == "string" then
        local itemID, itemBonus = str_match(itemData, ITEM_LOG_FORMAT)
        itemID = tonumber(itemID)
        if itemBonus and itemBonus ~= "" then
            itemBonus = format(ITEM_LINK_FORMT, itemID, itemBonus or "")
        else
            itemBonus = nil
        end
        ItemLinkCache[itemData] = {itemID, itemBonus}
        return itemID, itemBonus
    else
        return itemData
    end
end

-- ########################################
-- Instance Reset
-- Events
--		"GGL_OnInstancesResetNewLockout"	(instanceLockEntry, instanceLocks)
--		"GGL_OnInstancesResetUpdate"	(instanceLocks, nextReleaseStartTime, nextReleaseEndTimes)
--      "GGL_OnInstancesResetEnterInstance"     (enterCacheEntry, enterCache)
-- ########################################
local RESET_PATTERN =_G.INSTANCE_RESET_SUCCESS:gsub("%%s", "(.+)")
local RESET_UPDATE_TICK = 1
local MAX_RESETS = 10
local LOCK_TIME = 3600 -- 1h
local CACHE_TIME = 3600*2 -- 2h
local activeResetTimer, resetTimerIsRunning = false, false
local nextReleaseStartTime, nextReleaseEndTime

local function InstanceResetUpdateCache()
    local enterCache = GGL.db.InstanceResets.enterCache
    local curTime = GGL.GetTime()

    for i = 1, #enterCache do
        if (curTime - enterCache[i].time) > CACHE_TIME then
            tblremove(enterCache, i)
            return InstanceResetUpdateCache()
        end
    end
end

local function InstanceReset_OnUpdate(self, e)
    if not activeResetTimer then
        resetTimerIsRunning = false
        return
    end

	if nextReleaseEndTime <= GGL.GetTime() then
		GGL:InstanceResets_Update()
    end
    C_TimerAfter(RESET_UPDATE_TICK, InstanceReset_OnUpdate)
end

local function AddInstanceReset(zone)
    local instanceLocks = GGL.db.InstanceResets.locks
    local enterCache = GGL.db.InstanceResets.enterCache

    local enterTab

    -- check for enterTime
    for i = 1, #enterCache do
        if strfind(zone, enterCache[i].mapName) then
            enterTab = tblremove(enterCache, i)
            break
        end
    end

	instanceLocks[#instanceLocks + 1] = {
        zone = zone,
        enterTime = GGL.GetTime(),
		startTime = GGL.GetTime(),
        endTime = GGL.GetTime() + LOCK_TIME,
    }
    if enterTab then
        instanceLocks[#instanceLocks].mapID = enterTab.mapID
        instanceLocks[#instanceLocks].enterTime = enterTab.time or GGL.GetTime()
    end

    GGL:FireCallbackEvent("GGL_OnInstancesResetNewLockout", nil, instanceLocks[#instanceLocks], instanceLocks)
	GGL:InstanceResets_Update()
end

local function InstanceResets_OnMapChange(self, mapID, mapName, mapTypeFlag, difficultyID)
    if mapTypeFlag ~= GGL.MAP_TYPE_FLAGS.DUNGEON and mapTypeFlag ~= GGL.MAP_TYPE_FLAGS.RAID then return end

    InstanceResetUpdateCache()

    local enterCache = GGL.db.InstanceResets.enterCache

    for i = 1, #enterCache do
        if enterCache[i].mapID == mapID then
            return
        end
    end

    enterCache[#enterCache + 1] = {
        time = GGL.GetTime(),
        mapID = mapID,
        mapName = mapName,
        difficultyID = difficultyID
    }

    GGL:FireCallbackEvent("GGL_OnInstancesResetEnterInstance", nil, enterCache[#enterCache], enterCache)
end

function GGL:CHAT_MSG_SYSTEM(msg)
	local zone = strmatch(msg, RESET_PATTERN)
	if zone then
		AddInstanceReset(zone)
	end
end

GGL:RegisterEvent("CHAT_MSG_SYSTEM")
GGL:RegisterCallbackEvent("GGL_OnMapChange", InstanceResets_OnMapChange)

--- Updates all resets
function GGL:InstanceResets_Update()
	local instanceLocks = GGL.db.InstanceResets.locks
	if #instanceLocks <= 0 then
		activeResetTimer = false
		GGL:FireCallbackEvent("GGL_OnInstancesResetUpdate", nil, instanceLocks, nil, nil)
		return
	else
		for i = 1, #instanceLocks do
			if instanceLocks[i].endTime <= GGL.GetTime() then
				local remove = tblremove(instanceLocks, i)
                nextReleaseEndTime = nil
                GGL:FireCallbackEvent("GGL_OnInstancesResetLockRemoves", nil, remove)
				GGL:InstanceResets_Update()
				return
			end
			if not nextReleaseEndTime or nextReleaseEndTime > instanceLocks[i].endTime then
				nextReleaseStartTime = instanceLocks[i].startTime
				nextReleaseEndTime = instanceLocks[i].endTime
			end
		end
        activeResetTimer = true
        if not resetTimerIsRunning then
            resetTimerIsRunning = true
            C_TimerAfter(RESET_UPDATE_TICK, InstanceReset_OnUpdate)
        end
	end
	GGL:FireCallbackEvent("GGL_OnInstancesResetUpdate", nil, instanceLocks, nextReleaseStartTime, nextReleaseEndTime)
end

--- Gets a table with alle entered instances since x
-- @param       since               <number:time> (optional) get all instance enters since that time
-- @param       addCurrentZone      <bool> adds the current zone when since is set
function GGL:InstanceResets_GetInstanceEnterTimes(since, addCurrentZone)
    if since then
        local enterCache = GGL.db.InstanceResets.enterCache
        local curZone = GGL:GetCurrentMapID()
        local ret = {}

        for i = 1, #enterCache do
            if (enterCache[i].time >= since) or (curZone and enterCache[i].mapID == curZone) then
                ret[#ret+1] = enterCache[i]
            end
        end

        return ret
    else
        return GGL.db.InstanceResets.enterCache
    end
end

--- Returns the InstanceResetTable with startTime since x
-- @param       since       <number:time> (optional) get all instance enters since that time
function GGL:InstanceResets_GetInstanceResetTimes(since)
    if since then
        local instanceLocks = GGL.db.InstanceResets.locks
        local ret = {}

        for i = 1, #instanceLocks do
            if instanceLocks[i].startTime >= since then
                ret[#ret+1] = instanceLocks[i]
            end
        end

        return ret
    else
        return GGL.db.InstanceResets.locks
    end
end

-- ########################################
-- Island Expedition
-- Events
--		"GGL_OnIslandExpeditionFinish"	(mapID, difficultyID, isInParty, numPartyMember)
-- ########################################
GGL:RegisterEvent("ISLAND_COMPLETED")

function GGL:ISLAND_COMPLETED()
	local numPartyMember = GetNumGroupMembers(LE_PARTY_CATEGORY_INSTANCE)
    local mapID, mapName, mapTypeFlag, difficultyID = GGL:GetZoneData()
    local isInParty = numPartyMember and numPartyMember > 1

    GGL:FireCallbackEvent("GGL_OnIslandExpeditionFinish", nil, mapID, difficultyID, isInParty, isInParty and numPartyMember or nil)
end

-- ########################################
-- Kill log
-- Events
--      "GGL_OnPlayerKillNpc" (npcID, npcName)
-- #########################################
local function KillLog_OnNewKill(npcID, npcName)
    GGL:FireCallbackEvent("GGL_OnPlayerKillNpc", nil, npcID, npcName)
end

-- ########################################
-- Combat Log events
-- ########################################
local DAMAGE_EVENTS = {} -- DAMAGE_EVENTS[subEvent] = true

-- https://wow.gamepedia.com/UnitFlag
local COMBATLOG_OBJECT_REACTION_HOSTILE = 0x00000040
local COMBATLOG_OBJECT_TYPE_PLAYER = 0x00000400

-- savers
local KillLogValidGuidSaver = {}

-- https://wowpedia.fandom.com/wiki/COMBAT_LOG_EVENT
local function CLEU_Event(timestamp, subEvent, hideCaster, sourceGUID, sourceName, sourceFlags, sourceRaidFlags, destGUID, destName, destFlags, destRaidFlags, ...)
    if DAMAGE_EVENTS[subEvent] and sourceGUID == PlayerGUID then
        -- filter out players
        if bit_band(destFlags, COMBATLOG_OBJECT_TYPE_PLAYER) == 0 then -- bit_band(destFlags, COMBATLOG_OBJECT_REACTION_HOSTILE) ~= 0 / only aggro mobs
            local spellID, spellName, spellSchool, suf1, suf2 = ...
            -- check for overkill / maybe add a tag check?
            if suf2 and suf2 > 0 then
                KillLogValidGuidSaver[destGUID] = true
            end
        end
    elseif subEvent == "UNIT_DIED" and destFlags then
        -- filter out player deads
        if KillLogValidGuidSaver[destGUID] then
            KillLog_OnNewKill(GGL.GetNpcIDFromGuid(destGUID), destName)
            KillLogValidGuidSaver[destGUID] = nil
        end
    end
end

local combatLogFrame = CreateFrame("FRAME")
combatLogFrame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
combatLogFrame:SetScript("OnEvent", function(self, event, ...) CLEU_Event(CombatLogGetCurrentEventInfo()) end)

local CLEU_PREFIXES = {"SWING", "RANGE", "SPELL", "SPELL_PERIODIC", "SPELL_BUILDING"} -- ENVIRONMENTAL is not needed
local CLEU_DAMAGE_SUFFIXES = {"_DAMAGE", "_INSTAKILL"}
-- build DAMAGE_EVENTS
for pref = 1, #CLEU_PREFIXES do for suf = 1, #CLEU_DAMAGE_SUFFIXES do DAMAGE_EVENTS[ CLEU_PREFIXES[pref]..CLEU_DAMAGE_SUFFIXES[suf] ] = true end end