local AddonName, GGL = ...
local Templates = GGL.Templates
local InviteHandler = GGL.Invite

local L = GGL.Locale

local Invite, InviteHelper = Templates:Register("Invite")

-- lua
local tbl_remove = table.remove
local tonumber = tonumber
local strsub, match = string.sub, string.match

-- wow

-- ##################
-- local
-- ##################
local InviteList = {}

local INVITE_KEY_LIST = GGL.INVITE_KEY_LIST

-- ##################
-- static
-- ##################
local Static = {}
Templates.Invite = Static

-- ["type"] = "TemplateName"
local SessionTypes = {
    ["M"] = "Multibox",
    ["C"] = "Challenge",
}

--- Get a inviteTemplate from a given sessionID
-- @param       sessionID       <string>
-- @param       hostPlayer      <playerTemplate> if nil the player is used
-- @return inviteTemplate
function Static:Get(sessionID, hostPlayer)
    if not sessionID then return end
    if InviteList[sessionID] then
        return InviteList[sessionID]
    end

    local sType = SessionTypes[strsub(sessionID, 1, 1)]

    -- session type not found
    if not sType then return end

    -- register new session
    local session, isNewSession = GGL.Templates[sType]:GetSession(GGL.sessionDb.SessionRegister[sType], sessionID)

    -- check host data
    if not isNewSession and not session:IsValidHost(hostPlayer) then return end

    -- register new inviteTemplate
    InviteList[sessionID] = Templates:New("Invite", session, hostPlayer)
    InviteList[sessionID]:SetHost(hostPlayer)

    return InviteList[sessionID]
end

--- Creates a new inviteTemplate
-- @param       sessionType     <string/SessionTypes>
-- @param       hostPlayer      <playerTemplate> if nil the player is used
-- @return inviteTemplate
function Static:GetNew(sessionType, hostPlayer)
    if not sessionType or not SessionTypes[sessionType] then return end

    local sType = SessionTypes[sessionType]

    -- register new session
    local session = GGL.Templates[sType]:GetSession(GGL.sessionDb.SessionRegister[sType])

    -- get sessionID
    local sessionID = session:GetSessionID()

    -- register new inviteTemplate
    InviteList[sessionID] = Templates:New("Invite", session, hostPlayer)

    return InviteList[sessionID]
end

-- ##################
-- template

-- EVENTS
-- NewPlayerAdded(self, playerTemplate)     -- New player added into the list
-- PlayerRemoved(self, playerTemplate, inviteState)     -- Player removed from list
-- PlayerInviteStateUpdate(self, playerTemplate, inviteState)       -- Update invite state
-- InviteSend(self, playerTemplate)     -- Fires when a invite is send
-- OnDescriptionChange(self, description)     -- Fires when the description change
-- HostStartsSession(self, session, duration, endTime, settingsString)      -- Host starts the session
-- HostEndsSession(self, session)
-- ##################

local function Session_SetDataRefresh(self)
    -- update settins string
    self.inviteTemplate:GetSettings(true)

    self.inviteTemplate:SendNewSettings()
end

--- Invite Constructor
-- @param       session     <sessionTemplate>
-- @param       hostPlayer  Sets the host, nil/false means player is host
function Invite:Constructor(session, hostPlayer)
    if not session or not session.GetSessionID then return end
    local sessionID = session:GetSessionID()

    -- check if session is already registered
    if InviteList[sessionID] then
        return InviteList[sessionID]
    end

    -- add own callback handler
    GGL:AddCallbackHandler(self)

    -- set
    self.session = session
    self:SetHost(hostPlayer or session.selfPlayer)
    self.playerList = {}
    self.dbPlayerList = {}
    self.playerInviteID = {}

    -- Register Invite for handler
    InviteHandler:RegisterInviteTemplate(self)

    -- register playerUpdates
    self.OnPlayerUpdate = function(playerTemplate)
        if not playerTemplate.isOnline then
            self:SetPlayerInviteState(playerTemplate, INVITE_KEY_LIST.DECLINED)
        end
    end
    GGL.Templates.BNWoWPlayer:RegisterCallbackEvent("PlayerTemplate_PlayerUpdate", self.OnPlayerUpdate)

    InviteList[sessionID] = self
end


--- Close the session and invite
function Invite:Close()
    if self.isClosed then return end
    self.isClosed = true
    -- close auto invites
    self:OpenBattleNetInvites(false)
    -- end session
    self:EndSession()
    -- Unregister Invite from handler
    InviteHandler:UnregisterInviteTemplate(self)
    -- send event
    self:FireCallbackEvent("SessionTemplate_Invite_TemplateClose", self)
    -- unregister all events
    self:UnregisterAllCallbackEvents()
    -- unregister player updates
    GGL.Templates.BNWoWPlayer:UnregisterCallbackEvent("PlayerTemplate_PlayerUpdate", self.OnPlayerUpdate)
    -- close session
    self.session:Close()
end

--- Stops the session and invite
function Invite:Stop()
    -- close auto invites
    self:OpenBattleNetInvites(false)
    -- send out stop
    self:StopSession(true)

    -- only resume when session is running and not closed
    if self.isClosed or not self.sessionIsRunning then return end
    -- Unregister Invite from handler
    InviteHandler:UnregisterInviteTemplate(self)
    -- send event
    self:FireCallbackEvent("SessionTemplate_Invite_TemplateStop", self)
    -- unregister all events
    self:UnregisterAllCallbackEvents()
    -- stop session
    if not self.session:IsStopped() then
        self.session:Stop()
    end
end

--- Returns if player is host
-- @return playerIsHost
function Invite:PlayerIsHost()
    return self.playerIsHost
end

--- Sets a db list
-- @param       dbPlayerList    <table>
-- @param       load            <bool>  loads the db list into the template, this replace all existing
-- @format  [playerSyncID] = inviteState<GGL.INVITE_KEY_LIST>
function Invite:SetPlayerDBList(dbPlayerList, load)
    self.dbPlayerList = dbPlayerList or {}

    if load then
        wipe(self.playerList)
        for playerSyncID, inviteState in pairs(dbPlayerList) do
            self:AddPlayer(Templates.Sync.GetPlayerFromSync(playerSyncID))
        end
    end
end

--- Sets the description for the invite
-- @param       description     <string>
function Invite:SetDescription(description)
    if self.description ~= description then
        self.description = description
        self:FireCallbackEvent("SessionTemplate_OnDescriptionChange", self, description)
        self:SendNewDescription()
    else
        self.description = description
    end
end

--- Get the description of the invite/session
-- @return description
function Invite:GetDescription()
    return self.description or ""
end

--- Get the sessionTemplate of the inviteTemplate
-- @return sessionTemplate
function Invite:GetSession()
    return self.session
end

--- Get the sessionID
-- @return sessionID
function Invite:GetSessionID()
    return self.session:GetSessionID()
end

--- Get the sessionType
-- @return sessionType
function Invite:GetSessionType()
    return self.session:GetSessionType()
end

--- Sets the host for the invite and session
-- @param   playerTemplate      <playerTemplate>
function Invite:SetHost(playerTemplate)
    self.session:SetHostPlayer(playerTemplate, true)
    self.hostPlayer = self.session:GetHostPlayer()
    self.playerIsHost = self.session:PlayerIsHost()
end

--- Get the host playerTemplate
-- @return hostPlayerTemplate
function Invite:GetHost()
    return self.hostPlayer
end

--- Get session duration
-- @return sessionDuration
function Invite:GetDuration()
    return self.session:GetDuration()
end

--- Get endTime (from serverTime)
-- @return endTime
function Invite:GetEndTime()
    return self.session:GetEndTime()
end

--- Get the seralized settings string
-- @return settingsString
function Invite:GetSettings(reload)
    if reload or not self.settings then
        self.settings = GGL:Serialize(self.session.db.setSave)
    end
    return self.settings
end

--- Get own invite state, for host its always INVITE_KEY_LIST.ACCEPTED
-- @return inviteState
function Invite:GetOwnInviteState()
    if self.playerIsHost then
        return INVITE_KEY_LIST.ACCEPTED
    else
        return self.inviteState or INVITE_KEY_LIST.NONE
    end
end

--- checks if a playerTemplate is the host
-- @param       playerTemplate      <playerTemplate>
function Invite:PlayerTemplateIsHost(playerTemplate)
    if not playerTemplate then return end

    if playerTemplate == self.hostPlayer then
        return true
    elseif self.session:IsValidHost(playerTemplate) then
        self:SetHost(playerTemplate)
        return true
    end
end

-- ##################
-- Client only
-- ##################

--- Starts the Session
function Invite:HostSendStartSession(duration, endTime, settingsString)
    if self.sessionIsRunning or self.playerIsHost or self:GetOwnInviteState() ~= INVITE_KEY_LIST.ACCEPTED then return end
    self.sessionIsRunning = true

    self:FireCallbackEvent("SessionTemplate_Invite_HostStartsSession", self, self.session, duration, endTime, settingsString)

    GGL:Print(format(L["Session from %s started."], self.hostPlayer.displayName))
end

--- Ends the Session
function Invite:HostSendEndSession()
    if self.playerIsHost or self:GetOwnInviteState() ~= INVITE_KEY_LIST.ACCEPTED then return end
    self.sessionIsRunning = nil

    self:FireCallbackEvent("SessionTemplate_Invite_HostEndsSession", self, self.session)

    self:Close()

    GGL:Print(format(L["Session from %s ended."], self.hostPlayer.displayName))
end

--- Ends the Session
function Invite:HostSendStopSession()
    if self.playerIsHost or self:GetOwnInviteState() ~= INVITE_KEY_LIST.ACCEPTED then return end
    self.sessionIsRunning = false

    self:FireCallbackEvent("SessionTemplate_Invite_HostStopsSession", self, self.session)

    self:Stop()

    GGL:Print(format(L["Session from %s stoped."], self.hostPlayer.displayName))
end

--- Delist the Session
function Invite:HostSendDelistSession()
    if self.playerIsHost then return end
    self.sessionIsRunning = nil

    self:FireCallbackEvent("SessionTemplate_Invite_HostDelistSession", self, self.session)

    self:Stop()

    GGL:Print(format(L["Session from %s delisted."], self.hostPlayer.displayName))
end

--- Set own inviteState, only works if player is not the host
-- @param       newState    <INVITE_KEY_LIST>
function Invite:SetOwnInviteState(newState)
    if self.playerIsHost then return end
    if self.inviteState == newState then return end

    self.inviteState = newState

    InviteHandler:SendInviteState(self, self.hostPlayer)
    self:FireCallbackEvent("SessionTemplate_Invite_PlayerInviteStateUpdate", self, self.session.selfPlayer, newState)
end

--- Send own inviteState to host, only works if player is not the host
-- @param       newState    <INVITE_KEY_LIST>
function Invite:SendOwnInviteState()
    if self.playerIsHost or not self.inviteState then return end

    InviteHandler:SendInviteState(self, self.hostPlayer)
end

-- ##################
-- HOST ONLY
-- ##################

--- Add a player into the invite List
--- This triggers no invite!!! For invite and add use Invite:InvitePlayer
-- @param       playerTemplate      <playerTemplate>
-- @return  playerAdded<bool>
function Invite:AddPlayer(playerTemplate)
    if self.isClosed or not self.playerIsHost or not playerTemplate then return end

    -- check if player exists
    if self.playerList[playerTemplate] then return end

    -- Add Player
    self.playerList[playerTemplate] = true

    if not self.dbPlayerList[self.session:GetPlayerSyncID(playerTemplate)] then
        self.dbPlayerList[self.session:GetPlayerSyncID(playerTemplate)] = INVITE_KEY_LIST.NONE
    end

    -- Fire event for new player
    self:FireCallbackEvent("SessionTemplate_Invite_NewPlayerAdded", self, playerTemplate)

    return true
end

--- Remove player from the invite List and delist/kick them
-- @param       playerTemplate      <playerTemplate>
-- @return  playerRemoved<bool>
function Invite:RemovePlayer(playerTemplate)
    if self.isClosed or not self.playerIsHost then return end

    -- check if player is found
    if not self.playerList[playerTemplate] then
        return false
    end

    -- remove player
    self.playerList[playerTemplate] = nil

    -- send player removed event
    self:FireCallbackEvent("SessionTemplate_Invite_PlayerRemoved", self, playerTemplate, self.dbPlayerList[self.session:GetPlayerSyncID(playerTemplate)])

    -- remove player from db
    self.dbPlayerList[self.session:GetPlayerSyncID(playerTemplate)] = nil

    -- send delist
    InviteHandler:SendDelist(self, playerTemplate)

    return true
end

--- Add a player into the invite List and invites them
-- @param       playerTemplate      <playerTemplate>
-- @param       sendInstant         <bool> send out invite without delay
function Invite:InvitePlayer(playerTemplate, sendInstant)
    if self.isClosed or not self.playerIsHost then return end
    -- check if player even has the addon
    if not playerTemplate:CanSendMessage() then
        if not playerTemplate.isPlayer then
            local a, b = playerTemplate:CanSendMessage()
            GGL:AddIntoErrorLog("Invite:InvitePlayer / CanSendMessage", {playerTemplate.syncID, b})
        end
        return
    end
    -- add the player
    self:AddPlayer(playerTemplate)

    -- send invite
    InviteHandler:SendInvite(self, playerTemplate, sendInstant)

    -- fire event
    self:FireCallbackEvent("SessionTemplate_Invite_InviteSend", self, playerTemplate)
end

--- Invite a player if the player is on the playerList
-- @param       playerTemplate      <playerTemplate>
-- @param       sendInstant         <bool> send out invite without delay
function Invite:InviteRegisteredPlayer(playerTemplate, sendInstant)
    if self.isClosed or not self.playerIsHost then return end

    -- player not found in invite list?
    if not self.bnetInvitesOpen and not self.playerList[playerTemplate] then return end

    -- resend invite
    self:InvitePlayer(playerTemplate, sendInstant)
end

--- Gets called on a invite request
--- This gets send out instant
-- @param       playerTemplate      <playerTemplate>
-- @param       inviteID            <number/string> GGL.LOAD_TIME from the rquesting player
function Invite:OnPlayerRequestInvite(playerTemplate, inviteID)
    if self.isClosed or not self.playerIsHost then return end

    -- only reset invite state if there is a new rquest
    if not inviteID or not self.playerInviteID[playerTemplate] or self.playerInviteID[playerTemplate] ~= inviteID then
        -- as the player requests an invite the invite state is unknown
        self:SetPlayerInviteState(playerTemplate, INVITE_KEY_LIST.INVITED)
        self.playerInviteID[playerTemplate] = inviteID
    end

    if not self.bnetInvitesOpen then return end

    self:InviteRegisteredPlayer(playerTemplate, true)
end

--- Sets the invite state for a player
-- @param       playerTemplate      <playerTemplate>
-- @param       newState            <INVITE_KEY_LIST>
function Invite:SetPlayerInviteState(playerTemplate, newState)
    -- check if player exists
    if self.isClosed or not self.playerList[playerTemplate] then return end

    newState = tonumber(newState)

    self:SendStartSession(playerTemplate, true)

    -- no change for state
    if self.dbPlayerList[self.session:GetPlayerSyncID(playerTemplate)] == newState then return end

    -- set new state and fire update
    self.dbPlayerList[self.session:GetPlayerSyncID(playerTemplate)] = newState
    self:FireCallbackEvent("SessionTemplate_Invite_PlayerInviteStateUpdate", self, playerTemplate, newState)
end

--- Gets the invite state for a player
-- @param       playerTemplate      <playerTemplate>
function Invite:GetPlayerInviteState(playerTemplate)
    -- check if player exists
    if not self.playerList[playerTemplate] then return end

    return self.dbPlayerList[self.session:GetPlayerSyncID(playerTemplate)]
end

--- Invites new players from BN-Friendlist as long its open
-- @param       state       <bool>
function Invite:OpenBattleNetInvites(state)
    if not self.playerIsHost then return end

    if state then
        self.bnetInvitesOpen = true
        -- add players into list
        for playerGUID, player in Templates.BNWoWPlayer:IterPlayerList() do
            if player.isOnline then
                self:InvitePlayer(player)
            end
        end
        -- register update event
        if not self.OnNewPlayerRegistered then
            self.OnNewPlayerRegistered = function(playerTemplate)
                self:InvitePlayer(playerTemplate)
            end
            GGL.Templates.BNWoWPlayer:RegisterCallbackEvent("PlayerTemplate_NewPlayerRegistered", self.OnNewPlayerRegistered)
        end
        self.session.db.battleNetInvitesOpen = true
    else
        self.bnetInvitesOpen = false
        -- unregister update event
        if self.OnNewPlayerRegistered then
            GGL.Templates.BNWoWPlayer:UnregisterCallbackEvent("PlayerTemplate_NewPlayerRegistered", self.OnNewPlayerRegistered)
            self.OnNewPlayerRegistered = nil
            self.session.db.battleNetInvitesOpen = nil
        end
    end
end

--- Check if invites are open
function Invite:GetBattleNetInvitesState()
    return self.OnNewPlayerRegistered and true or false
end

--- Force send new invites for all players in the list
function Invite:ResendInvites()
    if self.isClosed or not self.playerIsHost then return end

    for playerTemplate in pairs(self.playerList) do
        self:InvitePlayer(playerTemplate)
    end
end

--- Sends out new settings set
function Invite:SendNewSettings()
    if self.isClosed or not self.playerIsHost then return end

    for playerTemplate in pairs(self.playerList) do
        if self:GetPlayerInviteState(playerTemplate) == INVITE_KEY_LIST.ACCEPTED then
            InviteHandler:SendSettings(self, playerTemplate)
        end
    end
end

--- Sends out the new description
-- @param       playerTemplate      <playerTemplate>
function Invite:SendNewDescription(playerTemplate)
    if self.isClosed or not self.playerIsHost then return end

    if playerTemplate then
        InviteHandler:SendDescription(self, playerTemplate)
    else
        for playerTemplate in pairs(self.playerList) do
            self:SendNewDescription(playerTemplate)
        end
    end
end

--- Sends out the start message
-- @param       playerTemplate      <playerTemplate>
-- @param       sendInstant         <bool> send out invite instant
function Invite:SendStartSession(playerTemplate, sendInstant)
    if playerTemplate and self.isClosed or not self.playerIsHost or not self.sessionIsRunning then return end

    if self:GetPlayerInviteState(playerTemplate) == INVITE_KEY_LIST.ACCEPTED then
        InviteHandler:SendStart(self, playerTemplate, sendInstant)
    end
end

--- Starts the Session for all players that accepted
function Invite:StartSession()
    if self.isClosed or not self.playerIsHost then return end
    self.sessionIsRunning = true

    -- check if bNet invites are open but not send out
    if self.session.db.battleNetInvitesOpen and not self.OnNewPlayerRegistered then
        self:OpenBattleNetInvites(true)
    end

    -- session events
    self.session:RegisterCallbackEvent("SessionTemplate_SetDataRefresh", Session_SetDataRefresh)

    for player, inviteState in self.session:IterInvitedPlayers() do
        self:SendStartSession(player)
    end
end

--- Sends out the stop message
-- @param       playerTemplate      <playerTemplate>
function Invite:SendStopSession(playerTemplate, force)
    if not self.playerIsHost or (not force and not self.sessionIsRunning) then return end

    InviteHandler:SendStop(self, playerTemplate)
end

--- Stops the Session for all players
function Invite:StopSession(force)
    if self.playerIsHost then
        for playerTemplate in pairs(self.playerList) do
            self:SendStopSession(playerTemplate, force)
        end
    else
        self:SetOwnInviteState(INVITE_KEY_LIST.DECLINED)
    end

    wipe(self.playerInviteID)

    self.sessionIsRunning = false
end

--- Sends out the end message
-- @param       playerTemplate      <playerTemplate>
function Invite:SendEndSession(playerTemplate)
    if not self.playerIsHost then return end

    InviteHandler:SendEnd(self, playerTemplate)
end

--- Ends the Session for all players
function Invite:EndSession()
    if not self.playerIsHost then return end

    for playerTemplate in pairs(self.playerList) do
        self:SendEndSession(playerTemplate)
    end

    wipe(self.playerInviteID)

    self.sessionIsRunning = nil
end